@@ -13,15 +13,15 @@ use HTTP::Date;
 
 =head1 NAME
 
-FCGI::Buffer - Verify and Optimise FCGI Output
+FCGI::Buffer - Verify, Cache and Optimise FCGI Output
 
 =head1 VERSION
 
-Version 0.02
+Version 0.03
 
 =cut
 
-our $VERSION = '0.02';
+our $VERSION = '0.03';
 
 =head1 SYNOPSIS
 
@@ -94,48 +94,62 @@ Create an FCGI::Buffer object.  Do one of these for each FCGI::Accept.
 
 # FIXME: Call init() on any arguments that are given
 sub new {
-	my $proto = shift;
-	my $class = ref($proto) || $proto;
+	my $class = shift;
+
+	return unless($class);
 
 	my $buf = IO::String->new();
 	my $old_buf = select($buf);
 
 	my $rc = {
 		buf => $buf,
-		old_buf => $old_buf
+		old_buf => $old_buf,
+		generate_304 => 1,
+		generate_last_modified => 1,
+		compress_content => 1,
+		optimise_content => 0,
+		lint_content => 0,
 	};
-	$rc->{generate_etag} = 1;
-	$rc->{generate_304} = 1;
-	$rc->{generate_last_modified} = 1;
-	$rc->{compress_content} = 1;
-	$rc->{optimise_content} = 0;
-	$rc->{lint_content} = 0;
-	$rc->{o} = ();
+	# $rc->{o} = ();
+
+	if($ENV{'SERVER_PROTOCOL'} &&
+	  ($ENV{'SERVER_PROTOCOL'} eq 'HTTP/1.1')) {
+	  	$rc->{generate_etag} = 1;
+	} else {
+	  	$rc->{generate_etag} = 0;
+	}
 
 	return bless $rc, $class;
 }
 
 sub DESTROY {
+	if(defined($^V) && ($^V ge 'v5.14.0')) {
+		return if ${^GLOBAL_PHASE} eq 'DESTRUCT';	# >= 5.14.0 only
+	}
 	my $self = shift;
 
-	if($self->{logger}) {
-		# This will cause everything to get flushed and prevent
-		# outputs to the logger.  We need to do that now since
-		# if we leave it to Perl to delete later we may get
-		# a mesage that Log4Perl::init() hasn't been called
-		$self->{logger} = undef;
-	}
 	select($self->{old_buf});
 	if(!defined($self->{buf})) {
 		return;
 	}
-	$self->{pos} = $self->{buf}->getpos;
+	my $pos = $self->{buf}->getpos;
 	$self->{buf}->setpos(0);
 	my $buf;
-	read($self->{buf}, $buf, $self->{pos});
+	read($self->{buf}, $buf, $pos);
 	my $headers;
 	($headers, $self->{body}) = split /\r?\n\r?\n/, $buf, 2;
 
+	if($self->{'logger'}) {
+		if($ENV{'HTTP_IF_NONE_MATCH'}) {
+			$self->{logger}->debug("HTTP_IF_NONE_MATCH: $ENV{HTTP_IF_NONE_MATCH}");
+		}
+		if($ENV{'HTTP_IF_MODIFIED_SINCE'}) {
+			$self->{logger}->debug("HTTP_IF_MODIFIED_SINCE: $ENV{HTTP_IF_MODIFIED_SINCE}");
+		}
+		$self->{logger}->debug("Generate_etag = $self->{generate_etag}");
+		$self->{logger}->debug("Generate_304 = $self->{generate_304}");
+		$self->{logger}->debug("Generate_last_modified = $self->{generate_last_modified}");
+	}
 	unless($headers || $self->is_cached()) {
 		# There was no output
 		return;
@@ -161,7 +175,7 @@ sub DESTROY {
 	if(defined($self->{body}) && ($self->{body} eq '')) {
 		# E.g. if header of Location is given with no body, for
 		#	redirection
-		$self->{body} = undef;
+		delete $self->{body};
 		if($self->{cache}) {
 			# Don't try to retrieve it below from the cache
 			$self->{send_body} = 0;
@@ -185,7 +199,7 @@ sub DESTROY {
 
 				# If we're on http://www.example.com and have a link
 				# to http://www.example.com/foo/bar.htm, change the
-				# link to /foo.bar.htm - there's no need to include
+				# link to /foo/bar.htm - there's no need to include
 				# the site name in the link
 				unless(defined($self->{info})) {
 					if($self->{cache}) {
@@ -272,6 +286,10 @@ sub DESTROY {
 		$self->{status} = $1;
 	}
 
+	if($self->{'logger'}) {
+		$self->{'logger'}->debug("Initial status = $self->{status}");
+	}
+
 	# Generate the eTag before compressing, since the compressed data
 	# includes the mtime field which changes thus causing a different
 	# Etag to be generated
@@ -283,12 +301,11 @@ sub DESTROY {
 		require Encode;
 		$encode_loaded = 1;
 		$self->{etag} = '"' . Digest::MD5->new->add(Encode::encode_utf8($self->{body}))->hexdigest() . '"';
-		push @{$self->{o}}, "ETag: $self->{etag}";
 		if($ENV{'HTTP_IF_NONE_MATCH'} && $self->{generate_304} && ($self->{status} == 200)) {
 			if($self->{logger}) {
 				$self->{logger}->debug("Compare $ENV{HTTP_IF_NONE_MATCH} with $self->{etag}");
 			}
-			if($self->{etag}) {
+			if($ENV{'HTTP_IF_NONE_MATCH'} eq $self->{etag}) {
 				push @{$self->{o}}, "Status: 304 Not Modified";
 				$self->{send_body} = 0;
 				$self->{status} = 304;
@@ -304,6 +321,7 @@ sub DESTROY {
 
 	if((length($encoding) > 0) && defined($self->{body})) {
 		my $range = $ENV{'Range'} ? $ENV{'Range'} : $ENV{'HTTP_RANGE'};
+
 		if($range && !$self->{cache}) {
 			# TODO: Partials
 			if($range =~ /^bytes=(\d*)-(\d*)/) {
@@ -349,6 +367,14 @@ sub DESTROY {
 					$cache_hash = Storable::thaw($self->{cobject}->value());
 					$headers = $cache_hash->{'headers'};
 					@{$self->{o}} = ("X-FCGI-Buffer-$VERSION: Hit");
+					if($self->{info}) {
+						my $host_name = $self->{info}->host_name();
+						push @{$self->{o}}, "X-Cache: HIT from $host_name";
+						push @{$self->{o}}, "X-Cache-Lookup: HIT from $host_name";
+					} else {
+						push @{$self->{o}}, 'X-Cache: HIT';
+						push @{$self->{o}}, 'X-Cache-Lookup: HIT';
+					}
 				} else {
 					carp "Error retrieving data for key $key";
 				}
@@ -376,9 +402,13 @@ sub DESTROY {
 					foreach my $k (keys %{$cache_hash}) {
 						$self->{body} .= "\t$k\n";
 					}
-					warn($self->{body});
 					$self->{cache}->remove($key);
-					carp "Can't retrieve body for key $key";
+					if($self->{logger}) {
+						$self->{logger}->error("Can't retrieve body for key $key");
+					} else {
+						carp "Can't retrieve body for key $key";
+					}
+					warn($self->{body});
 					$self->{send_body} = 0;
 					$self->{status} = 500;
 				}
@@ -386,7 +416,7 @@ sub DESTROY {
 			if($self->{send_body} && $ENV{'SERVER_PROTOCOL'} &&
 			  ($ENV{'SERVER_PROTOCOL'} eq 'HTTP/1.1') &&
 			  ($self->{status} == 200)) {
-				if($ENV{'HTTP_IF_NONE_MATCH'}) {
+				if($ENV{'HTTP_IF_NONE_MATCH'} && $self->{generate_etag}) {
 					if(!defined($self->{etag})) {
 						unless($encode_loaded) {
 							require Encode;
@@ -394,7 +424,10 @@ sub DESTROY {
 						}
 						$self->{etag} = '"' . Digest::MD5->new->add(Encode::encode_utf8($self->{body}))->hexdigest() . '"';
 					}
-					if(($self->{etag} =~ /\Q$ENV{'HTTP_IF_NONE_MATCH'}\E/) && $self->{generate_304}) {
+					if($self->{'logger'} && $self->{generate_304}) {
+						$self->{logger}->debug("Compare etags $ENV{HTTP_IF_NONE_MATCH} and $self->{etag}");
+					}
+					if(($ENV{'HTTP_IF_NONE_MATCH'} eq $self->{etag}) && $self->{generate_304}) {
 						push @{$self->{o}}, "Status: 304 Not Modified";
 						$self->{status} = 304;
 						$self->{send_body} = 0;
@@ -413,6 +446,9 @@ sub DESTROY {
 							require Encode;
 							$encode_loaded = 1;
 						}
+						if($self->{generate_etag} && !defined($self->{etag}) && ((!defined($headers)) || ($headers !~ /^ETag: /m))) {
+							$self->{etag} = '"' . Digest::MD5->new->add(Encode::encode_utf8($self->{body}))->hexdigest() . '"';
+						}
 						my $nbody = Compress::Zlib::memGzip(\Encode::encode_utf8($self->{body}));
 						if(length($nbody) < length($self->{body})) {
 							$self->{body} = $nbody;
@@ -423,13 +459,15 @@ sub DESTROY {
 				}
 			}
 			my $cannot_304 = !$self->{generate_304};
-			if(defined($headers) && ($headers =~ /^ETag: "([a-z0-9]{32})"/m)) {
-				$self->{etag} = $1;
-			} else {
-				$self->{etag} = $cache_hash->{'etag'};
+			unless($self->{etag}) {
+				if(defined($headers) && ($headers =~ /^ETag: "([a-z0-9]{32})"/m)) {
+					$self->{etag} = $1;
+				} else {
+					$self->{etag} = $cache_hash->{'etag'};
+				}
 			}
 			if($ENV{'HTTP_IF_NONE_MATCH'} && $self->{send_body} && ($self->{status} != 304) && $self->{generate_304}) {
-				if(defined($self->{etag}) && ($self->{etag} =~ /\Q$ENV{'HTTP_IF_NONE_MATCH'}\E/) && ($self->{status} == 200)) {
+				if(defined($self->{etag}) && ($self->{etag} eq $ENV{'HTTP_IF_NONE_MATCH'}) && ($self->{status} == 200)) {
 					push @{$self->{o}}, "Status: 304 Not Modified";
 					$self->{send_body} = 0;
 					$self->{status} = 304;
@@ -439,16 +477,18 @@ sub DESTROY {
 				} else {
 					$cannot_304 = 1;
 				}
-			} elsif($self->{generate_etag} && defined($self->{etag}) && ((!defined($headers)) || ($headers !~ /^ETag: /m))) {
-				push @{$self->{o}}, "ETag: $self->{etag}";
 			}
 			if($self->{cobject}) {
-				if($ENV{'HTTP_IF_MODIFIED_SINCE'} && ($self->{status} != 304) && (!$cannot_304)) {
+				if($ENV{'HTTP_IF_MODIFIED_SINCE'} && ($self->{status} != 304) && !$cannot_304) {
 					$self->_check_modified_since({
 						since => $ENV{'HTTP_IF_MODIFIED_SINCE'},
 						modified => $self->{cobject}->created_at()
 					});
-				} elsif($self->{generate_last_modified}) {
+				}
+				if(($self->{status} == 200) && $self->{generate_last_modified}) {
+					if($self->{logger}) {
+						$self->{logger}->debug('Set Last-Modified to ' . HTTP::Date::time2str($self->{cobject}->created_at()));
+					}
 					push @{$self->{o}}, "Last-Modified: " . HTTP::Date::time2str($self->{cobject}->created_at());
 				}
 			}
@@ -461,11 +501,7 @@ sub DESTROY {
 					$self->{cache_age} = '10 minutes';
 				}
 				$cache_hash->{'body'} = $unzipped_body;
-				if($self->{body} && $self->{send_body}) {
-					my $body_length = length($self->{body});
-					push @{$self->{o}}, "Content-Length: $body_length";
-				}
-				if(scalar(@{$self->{o}})) {
+				if($self->{o} && scalar(@{$self->{o}})) {
 					# Remember, we're storing the UNzipped
 					# version in the cache
 					my $c;
@@ -504,14 +540,54 @@ sub DESTROY {
 					}
 				}
 			}
+			if($self->{info}) {
+				my $host_name = $self->{info}->host_name();
+				if(defined($self->{x_cache})) {
+					push @{$self->{o}}, 'X-Cache: ' . $self->{x_cache} . " from $host_name";
+				} else {
+					push @{$self->{o}}, "X-Cache: MISS from $host_name";
+				}
+				push @{$self->{o}}, "X-Cache-Lookup: MISS from $host_name";
+			} else {
+				if(defined($self->{x_cache})) {
+					push @{$self->{o}}, 'X-Cache: ' . $self->{x_cache};
+				} else {
+					push @{$self->{o}}, 'X-Cache: MISS';
+				}
+				push @{$self->{o}}, 'X-Cache-Lookup: MISS';
+			}
 			push @{$self->{o}}, "X-FCGI-Buffer-$VERSION: Miss";
 		}
 		# We don't need it any more, so give Perl a chance to
 		# tidy it up seeing as we're in the destructor
-		$self->{cache} = undef;
+		delete $self->{cache};
+	} elsif($self->{info}) {
+		my $host_name = $self->{info}->host_name();
+		push @{$self->{o}}, ("X-Cache: MISS from $host_name", "X-Cache-Lookup: MISS from $host_name");
+	} else {
+		push @{$self->{o}}, ('X-Cache: MISS', 'X-Cache-Lookup: MISS');
 	}
 
-	my $body_length = defined($self->{body}) ? length($self->{body}) : 0;
+	if($self->{generate_etag} && ((!defined($headers)) || ($headers !~ /^ETag: /m))) {
+		if(defined($self->{etag})) {
+			push @{$self->{o}}, "ETag: $self->{etag}";
+			if($self->{logger}) {
+				$self->{logger}->debug("Set ETag to $self->{etag}");
+			}
+		} elsif($self->{logger} && (($self->{status} == 200) || $self->{status} == 304) && !$self->is_cached()) {
+			$self->{logger}->warn("BUG: ETag not generated, status $self->{status}");
+		}
+	}
+
+	my $body_length;
+	if(defined($self->{body})) {
+		if(utf8::is_utf8($self->{body})) {
+			utf8::encode($self->{body});
+		}
+		$body_length = length($self->{body});
+	} else {
+		$body_length = 0;
+	}
 
 	if(defined($headers) && length($headers)) {
 		# Put the original headers first, then those generated within
@@ -534,8 +610,7 @@ sub DESTROY {
 	}
 
 	if($body_length && $self->{send_body}) {
-		push @{$self->{o}}, '';
-		push @{$self->{o}}, $self->{body};
+		push @{$self->{o}}, ('', $self->{body});
 	}
 
 	# XXXXXXXXXXXXXXXXXXXXXXX
@@ -546,6 +621,14 @@ sub DESTROY {
 		$SIG{__WARN__} = sub {
 			$wideCharWarningsIssued += "@_" =~ /Wide character in .../;
 			$widemess = "@_";
+			if($self->{logger}) {
+				$self->{logger}->fatal($widemess);
+				my $i = 1;
+				$self->{logger}->trace('Stack Trace');
+				while((my @call_details = (caller($i++)))) {
+					$self->{logger}->trace($call_details[1] . ':' . $call_details[2] . ' in function ' . $call_details[3]);
+				}
+			}
 			CORE::warn(@_);     # call the builtin warn as usual
 		};
 
@@ -573,6 +656,11 @@ sub DESTROY {
 
 sub _check_modified_since {
 	my $self = shift;
+
+	if($self->{logger}) {
+		$self->{logger}->trace('In _check_modified_since');
+	}
+
 	if(!$self->{generate_304}) {
 		return;
 	}
@@ -631,7 +719,7 @@ sub _optimise_content {
 	$self->{body} =~ s/\n\s+|\s+\n/\n/g;
 	$self->{body} =~ s/\t+/ /g;
 	$self->{body} =~ s/\s(\<.+?\>\s\<.+?\>)/$1/;
-	$self->{body} =~ s/(\<.+?\>\s\<.+?\>)\s/$1/g;
+	# $self->{body} =~ s/(\<.+?\>\s\<.+?\>)\s/$1/g;
 	$self->{body} =~ s/\<p\>\s/\<p\>/gi;
 	$self->{body} =~ s/\<\/p\>\s\<p\>/\<\/p\>\<p\>/gi;
 	$self->{body} =~ s/\<\/tr\>\s\<tr\>/\<\/tr\>\<tr\>/gi;
@@ -697,6 +785,7 @@ Set various options and override default values.
 	optimise_content => 0,	# optimise your program's HTML, CSS and JavaScript
 	cache => CHI->new(driver => 'File'),	# cache requests
 	cache_key => 'string',	# key for the cache
+	cache_age => '10 minutes',	# how long to store responses in the cache
 	logger => $self->{logger},
 	lint->content => 0,	# Pass through HTML::Lint
 	generate_304 => 1,	# Generate 304: Not modified
@@ -707,7 +796,12 @@ The cache_key should be a unique value dependent upon the values set by the
 browser.
 
 The cache object will be an object that understands get_object(),
-set(), remove() and created_at() messages, such as an L<CHI> object.
+set(), remove() and created_at() messages, such as an L<CHI> object. It is
+used as a server-side cache to reduce the need to rerun database accesses.
+
+Items stay in the server-side cache by default for 10 minutes.
+This can be overridden by the cache_control HTTP header in the request, and
+the default can be changed by the cache_age argument to init().
 
 Logger will be an object that understands debug() such as an L<Log::Log4perl>
 object.
@@ -757,13 +851,13 @@ sub init {
 	}
 
 	# Unsafe options - must be called before output has been started
-	$self->{pos} = $self->{buf}->getpos;
-	if($self->{pos} > 0) {
+	my $pos = $self->{buf}->getpos;
+	if($pos > 0) {
 		if(defined($self->{logger})) {
-			$self->{logger}->warn("Too late to call init, $self->{pos} characters have been printed");
+			$self->{logger}->warn("Too late to call init, $pos characters have been printed");
 		} else {
 			# Must do Carp::carp instead of carp for Test::Carp
-			Carp::carp "Too late to call init, $self->{pos} characters have been printed";
+			Carp::carp "Too late to call init, $pos characters have been printed";
 		}
 	}
 	if(defined($params{cache}) && $self->can_cache()) {
@@ -781,6 +875,7 @@ sub init {
 				}
 			}
 		}
+		$self->{cache_age} ||= $params{cache_age};
 		$self->{cache} = $params{cache};
 		if(defined($params{cache_key})) {
 			$self->{cache_key} = $params{cache_key};
@@ -819,7 +914,11 @@ Returns true if the server is allowed to store the results locally.
 sub can_cache {
 	my $self = shift;
 
+	if(defined($self->{x_cache})) {
+		return ($self->{x_cache} eq 'HIT');
+	}
 	if(defined($ENV{'NO_CACHE'}) || defined($ENV{'NO_STORE'})) {
+		$self->{x_cache} = 'MISS';
 		return 0;
 	}
 	if(defined($ENV{'HTTP_CACHE_CONTROL'})) {
@@ -829,10 +928,13 @@ sub can_cache {
 		}
 		if(($control eq 'no-store') ||
 		       ($control eq 'no-cache') ||
+		       ($control eq 'max-age=0') ||
 		       ($control eq 'private')) {
+			$self->{x_cache} = 'MISS';
 			return 0;
 		}
 	}
+	$self->{x_cache} = 'HIT';
 	return 1;
 }
 
@@ -901,7 +1003,7 @@ sub is_cached {
 		if($self->{logger}) {
 			$self->{logger}->warn('is_cached: object is in the cache but not the data');
 		}
-		$self->{cobject} = undef;
+		delete $self->{cobject};
 		return 0;
 	}
 
@@ -914,7 +1016,7 @@ sub is_cached {
 		}
 		# Can't determine the age. Play it safe an assume we're not
 		# cached
-		$self->{cobject} = undef;
+		delete $self->{cobject};
 		return 0;
 	}
 	if($age > $self->{cobject}->created_at()) {
@@ -922,10 +1024,11 @@ sub is_cached {
 		if($self->{logger}) {
 			$self->{logger}->debug('Script has been updated');
 		}
-		$self->{cobject} = undef;
+		delete $self->{cobject};
 		# Nothing will be in date and all new searches would miss
 		# anyway, so may as well clear it all
-		$self->{cache}->clear();
+		# FIXME: RT104471
+		# $self->{cache}->clear();
 		return 0;
 	}
 	if($self->{logger}) {
@@ -961,6 +1064,10 @@ sub _my_age {
 sub _should_gzip {
 	my $self = shift;
 
+	if(defined($self->{send_body}) && ($self->{send_body} == 0)) {
+		return '';
+	}
+
 	if($self->{compress_content} && ($ENV{'HTTP_ACCEPT_ENCODING'} || $ENV{'HTTP_TE'})) {
 		my $accept = lc($ENV{'HTTP_ACCEPT_ENCODING'} ? $ENV{'HTTP_ACCEPT_ENCODING'} : $ENV{'HTTP_TE'});
 		foreach my $encoding ('x-gzip', 'gzip') {
@@ -1032,7 +1139,7 @@ your bug as I make changes.
 
 =head1 SEE ALSO
 
-HTML::Packer, HTML::Lint
+CGI::Buffer, HTML::Packer, HTML::Lint
 
 =head1 SUPPORT
 
